#include "libgfl.h"
#include "farcolor.hpp"
#include "plugin.hpp"
#include "farkeys.hpp"
#include <stdio.h>

#if defined(__BORLANDC__) && !defined(NO_CRT)
extern "C"
{
  BOOL WINAPI _export LibMain(HINSTANCE hDLLInst, DWORD fdwReason, LPVOID lpvReserved)
  {
    (void) hDLLInst;
    (void) fdwReason;
    (void) lpvReserved;
    return TRUE;
  }
}
#endif

static PluginStartupInfo Info;
static FARSTANDARDFUNCTIONS FSF;
//static char PluginRootKey[80];
static char **Exts=NULL;
static DWORD ExtsNum=0;
/*static struct Options
{
  int ShowInViewer;
} Opt;*/

enum
{
  MTitle,
  MImageInfo,
};

const char *GetMsg(int MsgId)
{
  return(Info.GetMsg(Info.ModuleNumber,MsgId));
}

/*
HKEY CreateRegKey(void)
{
  HKEY hKey;
  DWORD Disposition;
  RegCreateKeyEx(HKEY_CURRENT_USER,PluginRootKey,0,NULL,0,KEY_WRITE,NULL,&hKey,&Disposition);
  return(hKey);
}

HKEY OpenRegKey(void)
{
  HKEY hKey;
  if (RegOpenKeyEx(HKEY_CURRENT_USER,PluginRootKey,0,KEY_QUERY_VALUE,&hKey)!=ERROR_SUCCESS)
    return(NULL);
  return(hKey);
}

void SetRegKey(const char *ValueName,DWORD ValueData)
{
  HKEY hKey=CreateRegKey();
  RegSetValueEx(hKey,ValueName,0,REG_DWORD,(BYTE *)&ValueData,sizeof(ValueData));
  RegCloseKey(hKey);
}

int GetRegKey(const char *ValueName,int *ValueData,DWORD Default)
{
  HKEY hKey=OpenRegKey();
  DWORD Type,Size=sizeof(*ValueData);
  int ExitCode=RegQueryValueEx(hKey,ValueName,0,&Type,(BYTE *)ValueData,&Size);
  RegCloseKey(hKey);
  if (hKey==NULL || ExitCode!=ERROR_SUCCESS)
  {
    *ValueData=Default;
    return(FALSE);
  }
  return(TRUE);
}
*/

static void GetDIBFromBitmap(GFL_BITMAP *bitmap,BITMAPINFOHEADER *bitmap_info,unsigned char **data)
{
  int bytes_per_line;

  *data=NULL;
  memset(bitmap_info,0,sizeof(BITMAPINFOHEADER));

  bitmap_info->biSize=sizeof(BITMAPINFOHEADER);
  bitmap_info->biWidth=bitmap->Width;
  bitmap_info->biHeight=bitmap->Height;
  bitmap_info->biPlanes=1;

  bytes_per_line=(bitmap->Width*3+3)&-4;
  bitmap_info->biClrUsed=0;
  bitmap_info->biBitCount=24;
  bitmap_info->biCompression=BI_RGB;
  bitmap_info->biSizeImage=bytes_per_line*bitmap->Height;
  bitmap_info->biClrImportant=0;

  *data=(unsigned char*)HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,bitmap_info->biSizeImage);

  if(*data) memcpy(*data,bitmap->Data,bitmap_info->biSizeImage);
  return;
}

static RECT RangingPic(RECT DCRect,GFL_BITMAP *RawPicture)
{
  float asp_dst=(float)(DCRect.right-DCRect.left)/(float)(DCRect.bottom-DCRect.top);
  float asp_src=(float)RawPicture->Width/(float)RawPicture->Height;

  int dst_w;
  int dst_h;

  if(asp_dst<asp_src)
  {
    dst_w=min(DCRect.right-DCRect.left,RawPicture->Width);
    dst_h=(int)(dst_w/asp_src);
  }
  else
  {
    dst_h=min(DCRect.bottom-DCRect.top,RawPicture->Height);
    dst_w=(int)(asp_src*dst_h);
  }

  RECT dest={DCRect.left,DCRect.top,dst_w,dst_h};
  return dest;
}

static bool CheckName(const char *AFileName)
{
  size_t flen=strlen(AFileName);
  for(DWORD i=0;i<ExtsNum&&Exts;i++)
    if(Exts[i]&&strlen(Exts[i])<flen)
      if(!FSF.LStricmp(AFileName+flen-strlen(Exts[i]),Exts[i])&&AFileName[flen-strlen(Exts[i])-1]=='.') return true;
  return false;
}

struct DialogData
{
  HWND FarWindow;
  RECT DrawRect;
  RECT GDIRect;
  char FileName[MAX_PATH];
  bool Redraw;
  bool Loaded;
  BITMAPINFOHEADER *BmpHeader;
  unsigned char *DibData;
  GFL_FILE_INFORMATION *pic_info;
  int Page;
};

static bool DrawImage(DialogData *data)
{
  bool result=false;
  RECT DCRect,RangedRect;
  GFL_BITMAP *RawPicture=NULL;
  data->DibData=NULL;
  {
    RECT rect;
    CONSOLE_SCREEN_BUFFER_INFO info;
    GetClientRect(data->FarWindow,&rect);
    GetConsoleScreenBufferInfo(GetStdHandle(STD_OUTPUT_HANDLE),&info);

    int dx=rect.right/(info.srWindow.Right-info.srWindow.Left);
    int dy=rect.bottom/(info.srWindow.Bottom-info.srWindow.Top);

    DCRect.left=dx*(data->DrawRect.left-info.srWindow.Left);
    DCRect.right=dx*(data->DrawRect.right+1-info.srWindow.Left);
    DCRect.top=dy*(data->DrawRect.top-info.srWindow.Top);
    DCRect.bottom=dy*(data->DrawRect.bottom+1-info.srWindow.Top);
  }
  {
    GFL_LOAD_PARAMS load_params;
    gflGetDefaultLoadParams(&load_params);
    load_params.Flags|=GFL_LOAD_SKIP_ALPHA;
    load_params.Origin=GFL_BOTTOM_LEFT;
    load_params.LinePadding=4;
    load_params.ImageWanted=data->Page-1;
    GFL_ERROR res=gflLoadBitmap(data->FileName,&RawPicture,&load_params,data->pic_info);
    if(res) RawPicture=NULL;
    if(RawPicture)
    {
      if(!gflChangeColorDepth(RawPicture,NULL,GFL_MODE_TO_BGR,GFL_MODE_NO_DITHER))
      {
        GFL_BITMAP *pic=NULL;
        RangedRect=RangingPic(DCRect,RawPicture);
        RangedRect.left+=(DCRect.right-DCRect.left-RangedRect.right)/2;
        RangedRect.top+=(DCRect.bottom-DCRect.top-RangedRect.bottom)/2;
        gflResize(RawPicture,&pic,RangedRect.right,RangedRect.bottom,0,0);
        if(pic)
        {
          GetDIBFromBitmap(pic,data->BmpHeader,&data->DibData);
          gflFreeBitmap(pic);
        }
      }
    }
  }
  if(RawPicture&&data->DibData)
  {
    result=true;
    data->GDIRect=RangedRect;
  }
  if(RawPicture) gflFreeBitmap(RawPicture);
  return result;
}

static void UpdateImage(HANDLE hDlg,DialogData *data)
{
  if(!data->DibData&&!data->Loaded)
  {
    if(DrawImage(data))
    {
      char string[512];
      char *types[]={"RGB","GREY","CMY","CMYK","YCBCR","YUV16","LAB","LOGLUV","LOGL"};
      FSF.sprintf(string,GetMsg(MImageInfo),data->pic_info->Width,data->pic_info->Height,data->pic_info->Xdpi,data->pic_info->Ydpi,data->Page,data->pic_info->NumberOfImages,types[data->pic_info->ColorModel],data->pic_info->CompressionDescription);
      Info.SendDlgMessage(hDlg,DM_SETTEXTPTR,2,(long)string);
      data->Loaded=true;
    }
  }
  if(!data->DibData||!data->Loaded)
  {
    Info.SendDlgMessage(hDlg,DM_CLOSE,-1,0);
    return;
  }
  HDC hDC=GetDC(data->FarWindow);
  StretchDIBits(hDC,data->GDIRect.left,data->GDIRect.top,data->GDIRect.right,data->GDIRect.bottom,0,0,data->GDIRect.right,data->GDIRect.bottom,data->DibData,(BITMAPINFO *)data->BmpHeader,DIB_RGB_COLORS,SRCCOPY);
  ReleaseDC(data->FarWindow,hDC);
}

static void FreeImage(DialogData *data)
{
  HeapFree(GetProcessHeap(),0,data->DibData);
  data->DibData=NULL;
  gflFreeFileInformation(data->pic_info);
}

static long WINAPI PicDialogProc(HANDLE hDlg,int Msg,int Param1,long Param2)
{
  DialogData *DlgParams=(DialogData *)Info.SendDlgMessage(hDlg,DM_GETDLGDATA,0,0);
  switch(Msg)
  {
    case DN_INITDIALOG:
      Info.SendDlgMessage(hDlg,DM_SETDLGDATA,0,Param2);
      break;
    case DN_CTLCOLORDLGITEM:
      return (Info.AdvControl(Info.ModuleNumber,ACTL_GETCOLOR,(void*)COL_VIEWERSTATUS)<<24)|(Info.AdvControl(Info.ModuleNumber,ACTL_GETCOLOR,(void*)COL_VIEWERSTATUS)<<16)|(Info.AdvControl(Info.ModuleNumber,ACTL_GETCOLOR,(void*)COL_VIEWERTEXT)<<8)|(Info.AdvControl(Info.ModuleNumber,ACTL_GETCOLOR,(void*)COL_VIEWERSTATUS));
    case DN_DRAWDIALOG:
      DlgParams->Redraw=true;
      break;
    case DN_ENTERIDLE:
      if(DlgParams->Redraw)
      {
        DlgParams->Redraw=false;
        UpdateImage(hDlg,DlgParams);
      }
      break;
    case 0x3FFF:
      if(Param1) UpdateImage(hDlg,DlgParams);
      break;
    case DN_KEY:
      switch(Param2)
      {
        case KEY_CTRLR:
          UpdateImage(hDlg,DlgParams);
          return TRUE;
        case KEY_ADD:
        case KEY_SUBTRACT:
          if(DlgParams->DibData)
          {
            int Pages=DlgParams->pic_info->NumberOfImages;
            FreeImage(DlgParams);
            DlgParams->Loaded=false;
            if(Param2==KEY_ADD) DlgParams->Page++;
            else DlgParams->Page--;
            if(DlgParams->Page<1) DlgParams->Page=Pages;
            if(DlgParams->Page>Pages) DlgParams->Page=1;
            UpdateImage(hDlg,DlgParams);
          }
          return TRUE;
      }
      break;
  }
  return Info.DefDlgProc(hDlg,Msg,Param1,Param2);
}

extern "C" BOOL WINAPI _export GetJiggyWithIt(const char *FileName)
{
  if(Info.AdvControl(Info.ModuleNumber,ACTL_CONSOLEMODE,(void*)FAR_CONSOLE_GET_MODE)!=FAR_CONSOLE_FULLSCREEN&&CheckName(FileName))
  {
    DialogData data;
    data.FarWindow=(HWND)Info.AdvControl(Info.ModuleNumber,ACTL_GETFARHWND,0);
    CONSOLE_SCREEN_BUFFER_INFO info;
    GetConsoleScreenBufferInfo(GetStdHandle(STD_OUTPUT_HANDLE),&info);
    data.DrawRect.left=0;
    data.DrawRect.top=1;
    data.DrawRect.right=info.dwSize.X-1;
    data.DrawRect.bottom=info.dwSize.Y-2;
    strcpy(data.FileName,FileName);
    data.Redraw=false;
    data.Loaded=false;
    BITMAPINFOHEADER BmpHeader;
    data.BmpHeader=&BmpHeader;
    data.DibData=NULL;
    GFL_FILE_INFORMATION pic_info;
    data.pic_info=&pic_info;
    data.Page=1;
    FarDialogItem DialogItems[3];
    memset(DialogItems,0,sizeof(DialogItems));
    DialogItems[0].Type=DI_EDIT;
    Info.AdvControl(Info.ModuleNumber,FCTL_GETPANELINFO,(void *)&info);
    DialogItems[0].X1=0; DialogItems[0].X2=info.dwSize.X-1;
    DialogItems[0].Y1=0; DialogItems[0].Y2=0;
    strcpy(DialogItems[0].Data,FileName);
    DialogItems[1].Type=DI_USERCONTROL; DialogItems[1].Focus=TRUE;
    DialogItems[1].X1=0; DialogItems[1].X2=info.dwSize.X-1;
    DialogItems[1].Y1=1; DialogItems[1].Y2=info.dwSize.Y-2;
    DialogItems[2].Type=DI_EDIT;
    DialogItems[2].X1=0; DialogItems[2].X2=info.dwSize.X-1;
    DialogItems[2].Y1=info.dwSize.Y-1;
    DialogItems[2].Flags=DIF_READONLY;
    unsigned int VBufSize=(info.dwSize.Y-2)*(info.dwSize.X);
    int color=Info.AdvControl(Info.ModuleNumber,ACTL_GETCOLOR,(void*)COL_VIEWERTEXT)&0xF0;
    color=color|(color>>4);
    CHAR_INFO *VirtualBuffer;
    VirtualBuffer=(CHAR_INFO *)HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,VBufSize*sizeof(CHAR_INFO));
    DialogItems[1].VBuf=VirtualBuffer;
    for(unsigned int i=0;i<VBufSize;i++)
    {
      VirtualBuffer[i].Char.AsciiChar='.';
      VirtualBuffer[i].Attributes=color;
    }

    Info.DialogEx(Info.ModuleNumber,0,0,info.dwSize.X-1,info.dwSize.Y-1,NULL,DialogItems,sizeof(DialogItems)/sizeof(DialogItems[0]),0,FDLG_SMALLDIALOG,PicDialogProc,(long)&data);

    if(VirtualBuffer) HeapFree(GetProcessHeap(),0,VirtualBuffer);
    if(data.DibData) FreeImage(&data);
    return TRUE;
  }
  return FALSE;
}

HANDLE WINAPI OpenPlugin(int OpenFrom,int Item)
{
  char FileName[MAX_PATH];
  if (OpenFrom==OPEN_EDITOR || OpenFrom==OPEN_VIEWER)
  {
    struct WindowInfo wi;
    wi.Pos=-1;
    Info.AdvControl(Info.ModuleNumber,ACTL_GETWINDOWINFO,(void *)&wi);
    strcpy(FileName,wi.Name);
  }
  else if (OpenFrom==OPEN_COMMANDLINE)
  {
    strcpy(FileName,(const char *)Item);
    FSF.Unquote(FileName);
  }
  else if (OpenFrom==OPEN_PLUGINSMENU)
  {
    struct PanelInfo pi;
    Info.Control(INVALID_HANDLE_VALUE,FCTL_GETPANELINFO,(void *)&pi);
    if (pi.Plugin || !pi.ItemsNumber)
      return INVALID_HANDLE_VALUE;
    strcpy(FileName,pi.PanelItems[pi.CurrentItem].FindData.cFileName);
  }
  else
    return INVALID_HANDLE_VALUE;
  GetJiggyWithIt(FileName);
  return INVALID_HANDLE_VALUE;
}

void WINAPI GetPluginInfo(struct PluginInfo *Info)
{
  static const char *MenuStrings[1];
  Info->StructSize=sizeof(*Info);
  Info->Flags=PF_VIEWER|PF_EDITOR;
  MenuStrings[0]=GetMsg(MTitle);
  Info->PluginMenuStrings=MenuStrings;
  Info->PluginMenuStringsNumber=/*Opt.ShowInViewer?*/1/*:0*/;
  Info->CommandPrefix="picviews";
  //Info->PluginConfigStrings=MenuStrings;
  //Info->PluginConfigStringsNumber=1;
}

static void SetDefaultExtentions()
{
  int number=gflGetNumberOfFormat();
  GFL_FORMAT_INFORMATION finfo;
  for(int i=0;i<number;i++)
  {
    gflGetFormatInformationByIndex(i,&finfo);
    for(DWORD j=0;j<finfo.NumberOfExtension;j++)
    {
      Exts=(char **)(((Exts)?HeapReAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,Exts,(ExtsNum+1)*sizeof(*Exts)):HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,(ExtsNum+1)*sizeof(*Exts))));
      Exts[ExtsNum]=(char *)HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,strlen(finfo.Extension[j])+1);
      strcpy(Exts[ExtsNum],finfo.Extension[j]);
      ExtsNum++;
    }
  }
}

void WINAPI SetStartupInfo(const struct PluginStartupInfo *Info)
{
  ::Info=*Info;
  ::FSF=*Info->FSF;
  ::Info.FSF=&::FSF;
  //FSF.sprintf(PluginRootKey,"%s\\PicViewSimple",::Info.RootKey);
  //GetRegKey("ShowInViewer",&Opt.ShowInViewer,1);
  gflLibraryInit();
  gflEnableLZW(GFL_TRUE);
  SetDefaultExtentions();
}

void WINAPI ExitFAR()
{
  gflLibraryExit();
  if(Exts)
  {
    for(DWORD i=0;i<ExtsNum;i++)
      if(Exts[i])
        HeapFree(GetProcessHeap(),0,Exts[i]);
    HeapFree(GetProcessHeap(),0,Exts);
  }
  Exts=NULL;
  ExtsNum=0;
}

/*int WINAPI Configure(int ItemNumber)
{
  (void)ItemNumber;
  FarDialogItem DialogItems[2];

  memset(DialogItems,0,sizeof(DialogItems));
  DialogItems[0].Type=DI_DOUBLEBOX;
  DialogItems[0].X1=3; DialogItems[0].X2=50;
  DialogItems[0].Y1=1; DialogItems[0].Y2=3;
  strcpy(DialogItems[0].Data,GetMsg(MTitle));
  DialogItems[1].Type=DI_CHECKBOX;
  DialogItems[1].X1=5; //DialogItems[1].X2=48;
  DialogItems[1].Y1=2; //DialogItems[1].Y2=2;
  DialogItems[1].Focus=TRUE;
  GetRegKey("AutomaticInViewer",&Opt.AutomaticInViewer,0);
  DialogItems[1].Selected=Opt.AutomaticInViewer;
  //DialogItems[1].DefaultButton = 1;
  strcpy(DialogItems[1].Data,GetMsg(MAutomaticInViewer));

  if(Info.Dialog(Info.ModuleNumber,-1,-1,53,5,NULL,DialogItems,sizeof(DialogItems)/sizeof(DialogItems[0])) != 1)
    return FALSE;

  Opt.AutomaticInViewer=DialogItems[1].Selected;
  SetRegKey("AutomaticInViewer",Opt.AutomaticInViewer);
  return TRUE;
}*/

#if defined(__GNUC__)

#ifdef __cplusplus
extern "C"{
#endif
  bool WINAPI DllMainCRTStartup(HANDLE hDll,DWORD dwReason,LPVOID lpReserved);
#ifdef __cplusplus
};
#endif

bool WINAPI DllMainCRTStartup(HANDLE hDll,DWORD dwReason,LPVOID lpReserved)
{
  (void)hDll;
  (void)dwReason;
  (void)lpReserved;
  return true;
}

#endif